/*************************************************

   Programa simulador grafico 2D de un robot movil
   con un sistema de navegacin reactiva

	por: Jesus M Gomez de Gabriel
	Malaga, Agosto de 1992.
	Dto. Ingenieria de Sistema y Automatica
	Universidad de MALAGA


*************************************************/

#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>

#include "matrices.h"
#define TRUE 1
#define FALSE 0

#define COLORVEHICULO 9
#define COLORMAPA  10
#define COLORSTEERING 6
#define COLORSCANS    8
#define COLORCORTASCAN 15
#define COLORNOCORTASCAN 4

#define RADIOROBOT 75
#define MAXSTEERING PI/3
#define ACELERACION 1
#define NUMSCANS 24
#define LENGTHSCANS 1000

#define PI 3.14159

#define NUMSEGMENTOS 9

struct { float x1,y1,x2,y2; }
	mapa[NUMSEGMENTOS]={{-500,-500 , 750, -500},
			{750,-500,750,1000},
			{750,1000,0,1000},
			{0,1000,0,500},
			{0,500,-500,500},
			{250,500,500,500},
			{500,500,500,750},
			{500,750,250,750},
			{250,750,250,500}
			};

struct { float x1,y1,x2,y2; }
	mapa2[NUMSEGMENTOS]={{-400,-300 , 100, -300},
			{-200,-100,0,-100},
			{0,-100,0,300},
			{200,-100,200,50},
			{500,0,300,500}};

struct { int valido; float radio, x,y;} scans[NUMSCANS];

int maxbox;	/* dimensin de la ventana */
float escala = 0.2;
struct { float x, y, o, s, v, a, mindist; } robot =
	{ 350,900,0.2,0.3,0,0,100};

float viewmatrix[3][3],robotmatrix[3][3];
float avoidx,avoidy;


int corta(float x1, float y1, float x2, float y2,
	float x3, float y3, float x4, float y4, float *x, float *y);
float distancia(float x1, float y1, float x2,float y2);



float distancia(float x1, float y1, float x2, float y2)
{
float dx,dy;
dx=x2-x1;
dy=y2-y1;
return sqrt(dx*dx+dy*dy);
}



int corta(float x1, float y1, float x2, float y2,
	float x3, float y3, float x4, float y4, float *x, float *y)
{
float dx1,dy1,dx2,dy2,aux,px,py;
float m,a,b,n;

if (x2<x1) { aux=x2; x2=x1; x1=aux; aux=y2; y2=y1; y1=aux; }
if (x4<x3) { aux=x4; x4=x3; x3=aux; aux=y4; y4=y3; y3=aux; }

dx1=x2-x1;
dy1=y2-y1;
dx2=x4-x3;
dy2=y4-y3;

if ((dx1*dx2)!=0)	/* ningn segmento vertical */
	{
	m=dy1/dx1;
	a=dy2/dx2;
	n=y1-x1*m;
	b=y3-x3*a;
	if ((m-a)==0) return FALSE;  /* son segmentos paralelos */
	px=(b-n)/(m-a);
	py=m*(px)+n;
	if ((px<x1) || (px>x2) || (px<x3) || (px>x4) ) return FALSE;
	}
	else if ((dx1==0)&&(dx2==0)) /* ambos verticales */
		return FALSE;
	else
		{
		if (dx1==0) {
			if (y2<y1) { aux=y1; y1=y2; y2=aux;}
			a=dy2/dx2;
			b=y3-x3*a;
			px=x1;
			py=a*px+b;
			if ((py<y1) || (py>y2) || (px<x3) || (px>x4)) return FALSE;
			}
		else {
			if (y4<y3) { aux=y3; y3=y4; y4=aux;}
			m=dy1/dx1;
			n=y1-x1*m;
			px=x3;
			py=m*px+n;
			if ((py<y3) || (py>y4) || (px<x1) || (px>x2) ) return FALSE;
			}
		}

*x=px;
*y=py;
return TRUE;
}

void LeeScans()
{
int n,k;
float x,y,x2,y2,alfa,radio;
robot.mindist=LENGTHSCANS;
for (n=0; n<NUMSCANS; n++)	/* cada uno de los rayos */
	{
	alfa=robot.o+(2*PI*n)/NUMSCANS;
	x2=robot.x+LENGTHSCANS*cos(alfa);
	y2=robot.y+LENGTHSCANS*sin(alfa);
	scans[n].valido=FALSE;
	for (k=0; k<NUMSEGMENTOS; k++)	/* con cada uno de los segmentos */
		if (corta(	robot.x,robot.y,x2,y2,
				mapa[k].x1,mapa[k].y1,mapa[k].x2,mapa[k].y2,
				&x,&y))
			{
			radio=distancia(robot.x,robot.y,x,y);
			if ((!scans[n].valido)||(radio<scans[n].radio))
				{
				/* circle(M_tx(viewmatrix,x,y),M_ty(viewmatrix,x,y),4); */
				scans[n].x=x;
				scans[n].y=y;
				scans[n].radio=radio;
				scans[n].valido=TRUE;
				if (radio<robot.mindist) robot.mindist=radio;
				}
			}
	}
/* if (robot.mindist<RADIOROBOT) putchar(7);
*/
}

void SumaScans()
{
float alfa,fuerza;
int n;
avoidx=0.1;
avoidy=0;
for (n=0; n<NUMSCANS; n++) if (scans[n].valido)
	{
	alfa=2*PI*n/NUMSCANS;
	fuerza=log((scans[n].radio-RADIOROBOT)/(LENGTHSCANS*2))*10;
	avoidx+=fuerza*cos(alfa);
	avoidy+=fuerza*sin(alfa);
	}
M_moveto(robotmatrix,0,0);
M_lineto(robotmatrix,avoidx,avoidy);
}


void PintaMapa()
{
int n;
setcolor(COLORMAPA);
for (n=0; n<NUMSEGMENTOS; n++)
	{
	M_moveto(viewmatrix,mapa[n].x1,mapa[n].y1);
	M_lineto(viewmatrix,mapa[n].x2,mapa[n].y2);
	}
}

void PintaScans()
{
int n;
for (n=0; n<NUMSCANS; n++)
	if (scans[n].valido)
		{
		setcolor(COLORCORTASCAN);
		M_moveto(viewmatrix,robot.x,robot.y);
		M_lineto(viewmatrix,scans[n].x,scans[n].y);
		}
		else
		{
		float alfa;
		alfa=2*PI*n/NUMSCANS;
		setcolor(COLORNOCORTASCAN);
		M_moveto(robotmatrix,0,0);
		M_lineto(robotmatrix,LENGTHSCANS*cos(alfa),LENGTHSCANS*sin(alfa));
		}

}

void PintaPuntosScans()
{
int n;
setcolor(COLORCORTASCAN);
for (n=0; n<NUMSCANS; n++) if (scans[n].valido)
	circle(M_tx(viewmatrix,scans[n].x,scans[n].y),M_ty(viewmatrix,scans[n].x,scans[n].y),4);
ellipse(M_tx(robotmatrix,0,0),M_ty(robotmatrix,0,0),0,360,viewmatrix[0][0]*robot.mindist,-viewmatrix[1][1]*robot.mindist);
}

void PintaRobot()
{
int cx,cy,n;
setcolor(COLORVEHICULO);
M_moveto(robotmatrix,-50,-50);
M_lineto(robotmatrix,-50, 50);
M_lineto(robotmatrix, 50, 50);
M_lineto(robotmatrix, 50,-50);
M_lineto(robotmatrix,-50,-50);
cx=robotmatrix[0][2];
cy=robotmatrix[1][2];
moveto(cx,cy);
M_lineto(robotmatrix,100,0);
setcolor(COLORSTEERING);
moveto(cx,cy);
M_lineto(robotmatrix,100*cos(robot.s),100*sin(robot.s));
}

void MueveRobot(float x, float y, float o, float s)
{
float auxmatrix[3][3];
robot.x=x;
robot.y=y;
robot.o=o;
robot.s=s;
   M_mkmatriz(auxmatrix,	cos(robot.o),	-sin(robot.o),	robot.x,
				sin(robot.o),	cos(robot.o),	robot.y);
   M_copia(viewmatrix,robotmatrix);
   M_multiplica(robotmatrix,auxmatrix);
}

void Replay()
{
int page=0;
FILE * frep;
if ((frep=fopen("replay.dat","r"))!=NULL)
	while (fscanf(frep,"%f %f %f %f\n",&robot.x,&robot.y,&robot.o,&robot.s)==4)
		{
		page=1-page;
		setactivepage(page);
		cleardevice();
		PintaMapa();
		MueveRobot(robot.x,robot.y,robot.o,robot.s);
		PintaRobot();
		setvisualpage(page);
		}
fclose(frep);
}

void ReplayTrack()
{
FILE * frep;
cleardevice();
PintaMapa();
if ((frep=fopen("replay.dat","r"))!=NULL)
	while (fscanf(frep,"%f %f %f %f\n",&robot.x,&robot.y,&robot.o,&robot.s)==4)
		{
		MueveRobot(robot.x,robot.y,robot.o,robot.s);
		PintaRobot();
		}
fclose(frep);
}

int main(void)
{
   float angulo,fuerza;
   FILE * freplay;
   /* int gdriver = DETECT, gmode, errorcode; */
   int gdriver = EGA, gmode= EGAHI, errorcode;
   int salir=FALSE,page=0;
   int	scanactivo=TRUE,
	mapaactivo=TRUE,
	replayactivo=FALSE,
	avoidanceactivo=TRUE;

   initgraph(&gdriver, &gmode, "");
   errorcode = graphresult();
   if (errorcode != grOk)  /* an error occurred */
   {
      printf("Graphics error: %s\n", grapherrormsg(errorcode));
      printf("Press any key to halt:");
      getch();
      exit(1);             /* return with error code */
   }
   M_mkmatriz(viewmatrix,	escala,	0,		getmaxx()/2,
				0,	-escala*0.75,	getmaxy()/2);
   MueveRobot(robot.x,robot.y,robot.o,robot.s);
   freplay=fopen("replay.dat","w");
   do { page=1-page;
	setactivepage(page);
	cleardevice();
	if (mapaactivo) PintaMapa();
	PintaRobot();
	if (scanactivo)	{ 	LeeScans();
				PintaPuntosScans();
				SumaScans();
			}
	setvisualpage(page);
	if (avoidanceactivo)
		{
		fuerza=sqrt(avoidx*avoidx+avoidy*avoidy);
		robot.s=atan2(avoidy,avoidx);
		robot.a=log((robot.mindist-RADIOROBOT)*0.01)*5;
		robot.v+=robot.a*0.1;
		if (robot.v<10) robot.v=10;
		}
	while (kbhit()) switch(getch())
		{
		case 'i': robot.v+=ACELERACION; break;
		case 'l': robot.s-=0.1; break;
		case 'j': robot.s+=0.1; break;
		case 'k': robot.v-=ACELERACION; break;
		case 's': if (scanactivo) scanactivo=FALSE; else scanactivo=TRUE; break;
		case 'm': if (mapaactivo) mapaactivo=FALSE; else mapaactivo=TRUE; break;
		case ' ': robot.v=0; break;
		case 'r': replayactivo=TRUE; salir=TRUE; break;
		case 'a': if (avoidanceactivo) avoidanceactivo=FALSE; else avoidanceactivo=TRUE; break;
		case 27: salir =TRUE; break;
		}
	if (robot.s<-MAXSTEERING) robot.s=-MAXSTEERING;
		else if (robot.s>MAXSTEERING) robot.s=MAXSTEERING;
	robot.o+=robot.s*robot.v*0.01;
	robot.x+=cos(robot.o)*robot.v;
	robot.y+=sin(robot.o)*robot.v;
	MueveRobot(robot.x,robot.y,robot.o,robot.s);
	fprintf(freplay,"%f %f %f %f\n",robot.x,robot.y,robot.o,robot.s);
   } while (!salir);
   fclose(freplay);
   if (replayactivo)
	{
	Replay();
	salir=FALSE;
	do { switch(getch())
		{
		case 'r': Replay(); break;
		case 't': ReplayTrack(); break;
		default : salir=TRUE; break;
		}
	    } while (!salir);
	}
   closegraph();
   return 0;
}
